"""
Instead of a function that returns, T | None, wrap it in a `Some`.
This is so that when you want to railroad function execution, you can do it safely
and without having to do endless checks.
"""

from typing import Type, TypeVar, Self, Callable
import ergoflo.result as result

T = TypeVar("T")
U = TypeVar("U")


class Some[T]:
    def __init__(self, val: T):
        self.val = val

    def __eq__(self, other) -> bool:
        return self.val == other.val

    def __repr__(self) -> str:  # pragma: no cover
        return f"Some({self.val})"

    def map(self, val: T) -> "Some[T]":
        return Some(val)

    def unwrap(self) -> T:
        return self.val

    def map_or(self, _: T, f: Callable[..., U]) -> "Some[U]":
        return Some(f())

    def or_val(self, _: T) -> Self:
        return self

    def then(self, f: Callable[..., U]) -> "Some[U]":
        return Some(f(self.val))

    def filter(self, f: Callable[..., bool]) -> "Some[T] | Type[Nothing]":
        if f(self.val):
            return self
        return Nothing

    def ok_or(self, _) -> result.Ok[T]:
        return result.Ok(self.val)


class Nothing:
    def __repr__(self) -> str:  # pragma: no cover
        return "Nothing"

    @classmethod
    def unwrap(cls):
        raise RuntimeError("Attempted to unwrap null value")

    @classmethod
    def map(cls, _) -> Type[Self]:
        return cls

    @classmethod
    def map_or(cls, default: T, _: Callable[..., U]) -> Some[T]:
        return Some(default)

    @classmethod
    def or_val(cls, v: T) -> Some[T]:
        return Some(v)

    @classmethod
    def then(cls, _: Callable[..., T]) -> Type[Self]:
        return cls

    @classmethod
    def filter(cls, _: Callable[..., bool]) -> Type[Self]:
        return cls

    @classmethod
    def ok_or(cls, err: Exception) -> result.Err:
        return result.Err(err)


Maybe = Some[T] | Nothing



from contextlib import nullcontext
import pytest
from ergoflo.result import Ok, Err, Result


def test_ok_val():
    assert Ok(2).then(lambda x: x * 2) == Ok(4)


def test_err_val():
    assert Err(RuntimeError("foo")).then(lambda x: x * 2) == Err(RuntimeError("foo"))


@pytest.mark.parametrize(
    "input, expected",
    [(Ok(2), Ok("2")), (Err(RuntimeError("oops")), Err(RuntimeError("oops")))],
)
def test_map_result(input, expected):
    assert input.map(str) == expected


@pytest.mark.parametrize(
    "input, expected", [(Ok(2), Ok(2)), (Err(RuntimeError("oops")), Ok(4))]
)
def test_result_or(input, expected):
    assert input.or_(4) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (Ok(Ok(2)), Ok(2)),
        (Ok(3), Ok(3)),
        (Ok(Err(RuntimeError("oops"))), Err(RuntimeError("oops"))),
        (Err(RuntimeError("Oops")), Err(RuntimeError("Oops"))),
    ],
)
def test_flatten(input, expected):
    assert input.flatten() == expected


@pytest.mark.parametrize(
    "input, raises",
    [
        (Err(RuntimeError("oops")), pytest.raises(RuntimeError)),
        (Ok(2), nullcontext()),
    ],
)
def test_unwrap(input, raises):
    with raises:
        input.unwrap()


def test_bigger_chain():
    class ParseIntError(Exception):
        pass

    def inner_func(x: str) -> Result[int]:
        try:
            return Ok(int(x))
        except Exception:
            return Err(ParseIntError())

    assert inner_func("a").then(lambda x: x * 2) == Err(ParseIntError())



"""
A `flow` is another name for a pipeline.
Its aim is to chain multiple functions together, called sequentially on a single input.
"""

from typing import Callable
from functools import reduce


def flow[T](*funcs: Callable[..., T]) -> Callable[..., T]:
    """Aka \"compose\".
    `flow` chains multiple functions together, where the output of one is the input to another.

    `flow(f1, f2, f3)(x)` is the same as `f3(f2(f1(x)))`

    `flow` is useful for functions using `Maybe` and `Result` so that you can quickly make composable \
    pipelines so that they "railroad" expectedly.
    """

    def curried(input_value):
        # Apply each function in sequence to the input
        return reduce(lambda result, func: func(result), funcs, input_value)

        # Get the return type of the last function in *funcs

    return curried



"""
The `Result` type is used for functions that are *fallible*.

Meaning for example in cases that you would do a try/except and just raise
the error all the way up, you would instead catch it, wrap it in an `Err` and return
that to "railroad" the rest of the execution.
"""

from typing import TypeVar, Self, Callable

T = TypeVar("T")
U = TypeVar("U")


class Ok[T]:
    __match_args__ = ("v",)

    def __init__(self, val: T):
        self.val = val

    def __eq__(self, other) -> bool:
        return self.val == other.val

    def __repr__(self) -> str:  # pragma: no cover
        return f"Ok({self.val})"

    def then(self, f: Callable[..., U]) -> "Ok[U]":
        return Ok(f(self.val))

    def map(self, f: Callable[..., U]) -> "Ok[U]":
        return Ok(f(self.val))

    def or_(self, _: T) -> Self:
        return self

    def flatten(self) -> "Result[T]":
        match self.val:
            case Ok() | Err():
                return self.val
            case _:
                return self

    def unwrap(self) -> T:
        return self.val


class Err:
    __match_args__ = ("e",)

    def __init__(self, err: Exception):
        self._type = type(err)
        self.args = err.args

    def __eq__(self, other) -> bool:
        return self.args == other.args and self._type == other._type

    def __repr__(self) -> str:  # pragma: no cover
        return f"Err({self._type}({self.args}))"

    def then(self, _: Callable[..., U]) -> Self:
        return self

    def map(self, _: Callable[..., U]) -> Self:
        return self

    def or_(self, val: T) -> Ok[T]:
        return Ok(val)

    def flatten(self) -> Self:
        return self

    def unwrap(self) -> None:
        raise self._type(self.args)


Result = Ok[T] | Err



import pytest
from ergoflo.option import Some, Nothing
from ergoflo.result import Err, Ok


def test_get_some_val():
    assert Some(3).unwrap() == 3


def test_unwrap_nothing():
    with pytest.raises(RuntimeError):
        Nothing.unwrap()


@pytest.mark.parametrize(
    "input,map_val,expected", [(Some(3), 0, Some(0)), (Nothing, 0, Nothing)]
)
def test_map_option(input, map_val, expected):
    assert input.map(map_val) == expected


@pytest.mark.parametrize(
    "input,map_val,expected",
    [(Some(3), (1, lambda: 0), Some(0)), (Nothing, (2, lambda: 0), Some(2))],
)
def test_map_or_else_option(input, map_val, expected):
    assert input.map_or(*map_val) == expected


@pytest.mark.parametrize(
    "input,val,expected", [(Some(3), 0, Some(3)), (Nothing, 3, Some(3))]
)
def test_or_val_option(input, val, expected):
    assert input.or_val(val) == expected


def test_big_option_chain():
    x = Some(3).or_val(2).map(100)
    assert x == Some(100)


def test_big_option_chain_with_lambda():
    x = Some(3).or_val(100).map_or(3, lambda: "r")
    assert x == Some("r")


@pytest.mark.parametrize("input,expected", [(Some(3), Some(6)), (Nothing, Nothing)])
def test_option_then(input, expected):
    assert input.then(lambda x: x * 2).then(lambda x: x) == expected


@pytest.mark.parametrize(
    "input,expected", [(Some(4), Some(4)), (Some(3), Nothing), (Nothing, Nothing)]
)
def test_option_filter(input, expected):
    def is_even(x: int):
        return x % 2 == 0

    assert input.filter(is_even) == expected


@pytest.mark.parametrize(
    "input,expected",
    [(Some(2), Ok(2)), (Nothing, Err(RuntimeError("oops")))],
)
def test_option_ok_or(input, expected):
    assert input.ok_or(RuntimeError("oops")) == expected



from ergoflo.flow import flow
from ergoflo import Ok, Result, Err


def _add_two(x: int) -> int:
    return x + 2


def _fail_func(_: int) -> Result[int]:
    return Err(RuntimeError("oops"))


def test_simple_flow():
    assert flow(_add_two, _add_two)(0) == 4


def test_fallible_flow():
    assert flow(_add_two, _fail_func)(3) == Err(RuntimeError("oops"))
    assert flow(_add_two, lambda i: Ok(i))(3) == Ok(5)
